@@ -13,13 +13,23 @@
 extern int	 journal_size;
 extern int	 journal_flags;
 extern char	*journal_device;
+extern char	*journal_location_string;
+
+/*
+ * Flags for check_plausibility()
+ */
+#define CHECK_BLOCK_DEV	0x0001
+#define CREATE_FILE	0x0002
+#define CHECK_FS_EXIST	0x0004
+#define VERBOSE_CREATE	0x0008
 
 #ifndef HAVE_STRCASECMP
 extern int strcasecmp (char *s1, char *s2);
 #endif
 extern char *get_progname(char *argv_zero);
-extern void proceed_question(void);
-extern void check_plausibility(const char *device);
+extern void proceed_question(int delay);
+extern int check_plausibility(const char *device, int flags,
+			      int *ret_is_dev);
 extern void parse_journal_opts(const char *opts);
 extern void check_mount(const char *device, int force, const char *type);
 extern unsigned int figure_journal_size(int size, ext2_filsys fs);
